from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import Optional

from compas_timber.planning import BuildingPlan

if TYPE_CHECKING:
    pass


class BuildingPlanModelContainer:
    """Container for a BuildingPlan and the design elements it references.

    Parameters
    ----------
    plan : BuildingPlan
        The building plan.
    elements : dict, optional
        A dictionary of design elements referenced by the plan. Map element GUIDs to elements.
    geometries : dict, optional
        A dictionary of geometries associated with the design elements. Map element GUIDs to geometries.
    """

    def __init__(self, plan: BuildingPlan, elements: Optional[Dict[str, Any]] = None, geometries: Optional[Dict[str, Any]] = None):
        self.plan = plan
        self.elements = elements or {}
        self.geometries = geometries or {}

    def add_element(self, element: Any, geometry: Any):
        """
        Add a design element and its associated geometry to the container.

        Parameters
        ----------
        element : Element or Beam
            The design element to add. Can be either a Beam or an Element.
        geometry : Mesh, Box, Cylinder, Sphere, or str
            The geometry associated with the design element. If str, it's the filepath to an OBJ file loaded in the application.

        """
        # TODO: if not self.plan.has_element(element.guid) raise error/warning
        self.elements[element.guid] = element
        self.geometries[element.guid] = geometry



from compas.geometry import Frame
from compas.tolerance import TOL

from compas_pb import pb_dump_bts
from compas_pb import pb_load_bts

from compas_timber.elements import Beam
from compas_timber.fabrication import JackRafterCut


def test_placeholder():
    beam = Beam(Frame.worldXY(), length=5.0, width=0.2, height=0.3, name="goddamn_beam")

    loaded_beam: Beam = pb_load_bts(pb_dump_bts(beam))

    assert loaded_beam.frame == Frame.worldXY()
    assert TOL.is_close(loaded_beam.length, beam.length)
    assert TOL.is_close(loaded_beam.width, beam.width)
    assert TOL.is_close(loaded_beam.height, beam.height)


def test_processing():
    processing = JackRafterCut("start", 11.2, 33.4, 10.0, 45, 34)

    loaded: JackRafterCut = pb_load_bts(pb_dump_bts(processing))

    assert loaded.PROCESSING_NAME == processing.PROCESSING_NAME
    assert TOL.is_close(processing.start_x, loaded.start_x)
    assert TOL.is_close(processing.start_y, loaded.start_y)
    assert TOL.is_close(processing.angle, loaded.angle)
    assert TOL.is_close(processing.inclination, loaded.inclination)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: compas_timber_pb/generated/building_plan.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'compas_timber_pb/generated/building_plan.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from compas_pb.generated import mesh_pb2 as compas__pb_dot_generated_dot_mesh__pb2
from compas_pb.generated import box_pb2 as compas__pb_dot_generated_dot_box__pb2
from compas_pb.generated import cylinder_pb2 as compas__pb_dot_generated_dot_cylinder__pb2
from compas_pb.generated import sphere_pb2 as compas__pb_dot_generated_dot_sphere__pb2
from compas_timber_pb.generated import elements_pb2 as compas__timber__pb_dot_generated_dot_elements__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.compas_timber_pb/generated/building_plan.proto\x12\x15\x63ompas_timber_pb.data\x1a\x1e\x63ompas_pb/generated/mesh.proto\x1a\x1d\x63ompas_pb/generated/box.proto\x1a\"compas_pb/generated/cylinder.proto\x1a compas_pb/generated/sphere.proto\x1a)compas_timber_pb/generated/elements.proto\"\x96\x01\n\x08StepData\x12\x0c\n\x04guid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x14\n\x0c\x65lementGuids\x18\x03 \x03(\t\x12\x13\n\x0b\x65lementHeld\x18\x04 \x03(\t\x12\x0f\n\x07isBuilt\x18\x05 \x01(\x08\x12\x11\n\tisPlanned\x18\x06 \x01(\x08\x12\r\n\x05\x61\x63tor\x18\x07 \x01(\x05\x12\x10\n\x08priority\x18\x08 \x01(\x05\"^\n\x10\x42uildingPlanData\x12\x0c\n\x04guid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12.\n\x05steps\x18\x03 \x03(\x0b\x32\x1f.compas_timber_pb.data.StepData\"y\n\x07\x45lement\x12/\n\x04\x62\x65\x61m\x18\x01 \x01(\x0b\x32\x1f.compas_timber_pb.data.BeamDataH\x00\x12\x35\n\x07\x65lement\x18\x02 \x01(\x0b\x32\".compas_timber_pb.data.ElementDataH\x00\x42\x06\n\x04kind\"\xe7\x01\n\x0f\x45lementGeometry\x12\x1a\n\x10\x66irebaseFilename\x18\x01 \x01(\tH\x00\x12(\n\x04mesh\x18\x02 \x01(\x0b\x32\x18.compas_pb.data.MeshDataH\x00\x12&\n\x03\x62ox\x18\x03 \x01(\x0b\x32\x17.compas_pb.data.BoxDataH\x00\x12\x30\n\x08\x63ylinder\x18\x04 \x01(\x0b\x32\x1c.compas_pb.data.CylinderDataH\x00\x12,\n\x06sphere\x18\x05 \x01(\x0b\x32\x1a.compas_pb.data.SphereDataH\x00\x42\x06\n\x04kind\"\xb5\x03\n\x1e\x42uildingPlanModelContainerData\x12\x35\n\x04plan\x18\x01 \x01(\x0b\x32\'.compas_timber_pb.data.BuildingPlanData\x12U\n\x08\x65lements\x18\x02 \x03(\x0b\x32\x43.compas_timber_pb.data.BuildingPlanModelContainerData.ElementsEntry\x12Y\n\ngeometries\x18\x03 \x03(\x0b\x32\x45.compas_timber_pb.data.BuildingPlanModelContainerData.GeometriesEntry\x1aO\n\rElementsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12-\n\x05value\x18\x02 \x01(\x0b\x32\x1e.compas_timber_pb.data.Element:\x02\x38\x01\x1aY\n\x0fGeometriesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x35\n\x05value\x18\x02 \x01(\x0b\x32&.compas_timber_pb.data.ElementGeometry:\x02\x38\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'compas_timber_pb.generated.building_plan_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_BUILDINGPLANMODELCONTAINERDATA_ELEMENTSENTRY']._loaded_options = None
  _globals['_BUILDINGPLANMODELCONTAINERDATA_ELEMENTSENTRY']._serialized_options = b'8\001'
  _globals['_BUILDINGPLANMODELCONTAINERDATA_GEOMETRIESENTRY']._loaded_options = None
  _globals['_BUILDINGPLANMODELCONTAINERDATA_GEOMETRIESENTRY']._serialized_options = b'8\001'
  _globals['_STEPDATA']._serialized_start=250
  _globals['_STEPDATA']._serialized_end=400
  _globals['_BUILDINGPLANDATA']._serialized_start=402
  _globals['_BUILDINGPLANDATA']._serialized_end=496
  _globals['_ELEMENT']._serialized_start=498
  _globals['_ELEMENT']._serialized_end=619
  _globals['_ELEMENTGEOMETRY']._serialized_start=622
  _globals['_ELEMENTGEOMETRY']._serialized_end=853
  _globals['_BUILDINGPLANMODELCONTAINERDATA']._serialized_start=856
  _globals['_BUILDINGPLANMODELCONTAINERDATA']._serialized_end=1293
  _globals['_BUILDINGPLANMODELCONTAINERDATA_ELEMENTSENTRY']._serialized_start=1123
  _globals['_BUILDINGPLANMODELCONTAINERDATA_ELEMENTSENTRY']._serialized_end=1202
  _globals['_BUILDINGPLANMODELCONTAINERDATA_GEOMETRIESENTRY']._serialized_start=1204
  _globals['_BUILDINGPLANMODELCONTAINERDATA_GEOMETRIESENTRY']._serialized_end=1293
# @@protoc_insertion_point(module_scope)



from compas.plugins import plugin


@plugin(category="factories", requires=["compas_pb"])
def register_serializers():
    import compas_timber_pb.data  # noqa: F401

    print("Discovered compas_pb plugin for compas_timber_pb")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: compas_timber_pb/generated/elements.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'compas_timber_pb/generated/elements.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from compas_pb.generated import frame_pb2 as compas__pb_dot_generated_dot_frame__pb2
from compas_pb.generated import mesh_pb2 as compas__pb_dot_generated_dot_mesh__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n)compas_timber_pb/generated/elements.proto\x12\x15\x63ompas_timber_pb.data\x1a\x1f\x63ompas_pb/generated/frame.proto\x1a\x1e\x63ompas_pb/generated/mesh.proto\"\x7f\n\x0b\x45lementData\x12\x0c\n\x04guid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12(\n\x05\x66rame\x18\x03 \x01(\x0b\x32\x19.compas_pb.data.FrameData\x12*\n\x08geometry\x18\x04 \x01(\x0b\x32\x18.compas_pb.data.MeshData\"\x7f\n\x08\x42\x65\x61mData\x12\x0c\n\x04guid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0e\n\x06length\x18\x03 \x01(\x02\x12\r\n\x05width\x18\x04 \x01(\x02\x12\x0e\n\x06height\x18\x05 \x01(\x02\x12(\n\x05\x66rame\x18\x06 \x01(\x0b\x32\x19.compas_pb.data.FrameDatab\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'compas_timber_pb.generated.elements_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_ELEMENTDATA']._serialized_start=133
  _globals['_ELEMENTDATA']._serialized_end=260
  _globals['_BEAMDATA']._serialized_start=262
  _globals['_BEAMDATA']._serialized_end=389
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: compas_timber_pb/generated/processing.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'compas_timber_pb/generated/processing.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from compas_pb.generated import message_pb2 as compas__pb_dot_generated_dot_message__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n+compas_timber_pb/generated/processing.proto\x12\x15\x63ompas_timber_pb.data\x1a!compas_pb/generated/message.proto\"\xbf\x01\n\x12\x42TLxProcessingData\x12\x0c\n\x04guid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x45\n\x06params\x18\x03 \x03(\x0b\x32\x35.compas_timber_pb.data.BTLxProcessingData.ParamsEntry\x1a\x46\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12&\n\x05value\x18\x02 \x01(\x0b\x32\x17.compas_pb.data.AnyData:\x02\x38\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'compas_timber_pb.generated.processing_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_BTLXPROCESSINGDATA_PARAMSENTRY']._loaded_options = None
  _globals['_BTLXPROCESSINGDATA_PARAMSENTRY']._serialized_options = b'8\001'
  _globals['_BTLXPROCESSINGDATA']._serialized_start=106
  _globals['_BTLXPROCESSINGDATA']._serialized_end=297
  _globals['_BTLXPROCESSINGDATA_PARAMSENTRY']._serialized_start=227
  _globals['_BTLXPROCESSINGDATA_PARAMSENTRY']._serialized_end=297
# @@protoc_insertion_point(module_scope)



from compas.geometry import Box
from compas.geometry import Cylinder
from compas.geometry import Sphere
from compas_model.elements import Element
from compas_pb.conversions import box_from_pb
from compas_pb.conversions import box_to_pb
from compas_pb.conversions import cylinder_from_pb
from compas_pb.conversions import cylinder_to_pb
from compas_pb.conversions import frame_from_pb
from compas_pb.conversions import frame_to_pb
from compas_pb.conversions import mesh_from_pb
from compas_pb.conversions import mesh_to_pb
from compas_pb.conversions import sphere_from_pb
from compas_pb.conversions import sphere_to_pb
from compas_pb.core import any_from_pb
from compas_pb.core import any_to_pb
from compas_pb.registry import pb_deserializer
from compas_pb.registry import pb_serializer
from compas_timber.elements import Beam
from compas_timber.fabrication import BTLxProcessing
from compas_timber.planning import BuildingPlan
from compas_timber.planning import Step

from compas_timber_pb.generated import building_plan_pb2
from compas_timber_pb.generated import elements_pb2
from compas_timber_pb.generated import processing_pb2
from compas_timber_pb.planning import BuildingPlanModelContainer


@pb_serializer(BuildingPlanModelContainer)
def buildingplanmodelcontainer_to_pb(obj: BuildingPlanModelContainer) -> building_plan_pb2.BuildingPlanModelContainerData:
    proto_data = building_plan_pb2.BuildingPlanModelContainerData()
    proto_data.plan.CopyFrom(buildingplan_to_pb(obj.plan))

    for guid, element in obj.elements.items():
        proto_element = building_plan_pb2.Element()
        if isinstance(element, Beam):
            proto_element.beam.CopyFrom(beam_to_pb(element))
        elif isinstance(element, Element):
            proto_element.element.CopyFrom(element_to_pb(element))
        proto_data.elements[guid].CopyFrom(proto_element)

    for guid, geometry in obj.geometries.items():
        proto_geometry = building_plan_pb2.ElementGeometry()
        if isinstance(geometry, str):
            proto_geometry.firebaseFilename = geometry
        elif isinstance(geometry, Box):
            proto_geometry.box.CopyFrom(box_to_pb(geometry))
        elif isinstance(geometry, Cylinder):
            proto_geometry.cylinder.CopyFrom(cylinder_to_pb(geometry))
        elif isinstance(geometry, Sphere):
            proto_geometry.sphere.CopyFrom(sphere_to_pb(geometry))
        elif hasattr(geometry, "to_mesh"):
            proto_geometry.mesh.CopyFrom(mesh_to_pb(geometry.to_mesh()))
        proto_data.geometries[guid].CopyFrom(proto_geometry)

    return proto_data


@pb_deserializer(building_plan_pb2.BuildingPlanModelContainerData)
def buildingplanmodelcontainer_from_pb(proto_data: building_plan_pb2.BuildingPlanModelContainerData) -> BuildingPlanModelContainer:
    plan = buildingplan_from_pb(proto_data.plan)
    elements = {}

    for guid, proto_element in proto_data.elements.items():
        if proto_element.HasField("beam"):
            elements[guid] = beam_from_pb(proto_element.beam)
        elif proto_element.HasField("element"):
            elements[guid] = element_from_pb(proto_element.element)

    geometries = {}
    for guid, proto_geometry in proto_data.geometries.items():
        if proto_geometry.HasField("firebaseFilename"):
            geometries[guid] = proto_geometry.firebaseFilename
        elif proto_geometry.HasField("box"):
            geometries[guid] = box_from_pb(proto_geometry.box)
        elif proto_geometry.HasField("cylinder"):
            geometries[guid] = cylinder_from_pb(proto_geometry.cylinder)
        elif proto_geometry.HasField("sphere"):
            geometries[guid] = sphere_from_pb(proto_geometry.sphere)
        elif proto_geometry.HasField("mesh"):
            geometries[guid] = mesh_from_pb(proto_geometry.mesh)

    return BuildingPlanModelContainer(plan, elements, geometries)


@pb_serializer(Beam)
def beam_to_pb(obj: Beam) -> elements_pb2.BeamData:
    """Convert a Beam object to a protobuf message.

    Parameters
    ----------
    obj : Beam
        The Beam object to convert.

    Returns
    -------
    elements_pb2.BeamData
        The protobuf message representation of the Beam.
    """
    proto_data = elements_pb2.BeamData()
    proto_data.guid = str(obj.guid)
    proto_data.name = obj.name
    proto_data.length = obj.length
    proto_data.width = obj.width
    proto_data.height = obj.height
    frame = frame_to_pb(obj.frame)

    proto_data.frame.CopyFrom(frame)
    return proto_data


@pb_deserializer(elements_pb2.BeamData)
def beam_from_pb(proto_data: elements_pb2.BeamData) -> Beam:
    """Convert a protobuf message to a Beam object.

    Parameters
    ----------
    proto_data : elements_pb2.BeamData
        The protobuf message to convert.

    Returns
    -------
    Beam
        The reconstructed Beam object.
    """
    frame = frame_from_pb(proto_data.frame)
    beam = Beam(
        frame=frame,
        length=proto_data.length,
        width=proto_data.width,
        height=proto_data.height,
        name=proto_data.name,
    )
    beam._guid = proto_data.guid
    return beam


@pb_serializer(Element)
def element_to_pb(obj: Element) -> elements_pb2.ElementData:
    """Convert an Element object to a protobuf message.

    Parameters
    ----------
    obj : Element
        The Element object to convert.

    Returns
    -------
    elements_pb2.ElementData
        The protobuf message representation of the Element.
    """
    proto_data = elements_pb2.ElementData()
    proto_data.guid = str(obj.guid)
    proto_data.name = obj.name

    frame = frame_to_pb(obj.frame)
    proto_data.frame.CopyFrom(frame)

    mesh = mesh_to_pb(obj.geometry)
    proto_data.geometry.CopyFrom(mesh)

    return proto_data


@pb_deserializer(elements_pb2.ElementData)
def element_from_pb(proto_data: elements_pb2.ElementData) -> Element:
    """Convert a protobuf message to an Element object.

    Parameters
    ----------
    proto_data : elements_pb2.ElementData
        The protobuf message to convert.

    Returns
    -------
    Element
        The reconstructed Element object.
    """
    frame = frame_from_pb(proto_data.frame)
    mesh = mesh_from_pb(proto_data.geometry)
    element = Element(frame=frame, name=proto_data.name, geometry=mesh)
    element._guid = proto_data.guid
    return element


@pb_serializer(BTLxProcessing)
def processing_to_pb(obj: BTLxProcessing) -> processing_pb2.BTLxProcessingData:
    """Convert a BTLxProcessing object to a protobuf message.

    Parameters
    ----------
    obj : BTLxProcessing
        The BTLxProcessing object to convert.

    Returns
    -------
    processing_pb2.BTLxProcessingData
        The protobuf message representation of the BTLxProcessing object.
    """
    proto_data = processing_pb2.BTLxProcessingData()
    proto_data.name = obj.PROCESSING_NAME
    proto_data.guid = str(obj.guid)

    for key, value in obj.__data__.items():
        data = any_to_pb(value)
        proto_data.params[key].CopyFrom(data)

    return proto_data


def _fetch_processing_cls_by_name(name: str):
    """Fetch the processing class by name.

    Parameters
    ----------
    name : str
        The name of the processing class to fetch.

    Returns
    -------
    type
        The processing class with the specified name.

    Raises
    ------
    ValueError
        If no processing class with the specified name is found.
    """
    for cls in BTLxProcessing.__subclasses__():
        if cls.PROCESSING_NAME == name:
            return cls
    raise ValueError(f"Processing class with name '{name}' not found.")


@pb_deserializer(processing_pb2.BTLxProcessingData)
def processing_from_pb(proto_data: processing_pb2.BTLxProcessingData) -> BTLxProcessing:
    """Convert a protobuf message to a BTLxProcessing object.

    Parameters
    ----------
    proto_data : processing_pb2.BTLxProcessingData
        The protobuf message to convert.

    Returns
    -------
    BTLxProcessing
        The reconstructed BTLxProcessing object.
    """

    cls = _fetch_processing_cls_by_name(proto_data.name)
    data_dict = {}
    for key, value in proto_data.params.items():
        data = any_from_pb(value)
        data_dict[key] = data
    instance = cls.__from_data__(data_dict)
    instance._guid = proto_data.guid  # TODO: should be able to send this to __from_data__, or is it __from_json__?
    return instance


@pb_serializer(BuildingPlan)
def buildingplan_to_pb(obj: BuildingPlan) -> building_plan_pb2.BuildingPlanData:
    """Convert a BuildingPlan object to a protobuf message.

    Parameters
    ----------
    obj : BuildingPlan
        The BuildingPlan object to convert.

    Returns
    -------
    building_plan_pb2.BuildingPlanData
        The protobuf message representation of the BuildingPlan.
    """
    proto_data = building_plan_pb2.BuildingPlanData()
    proto_data.guid = str(obj.guid)
    proto_data.name = obj.name

    for step in obj.steps:
        proto_step = building_plan_pb2.StepData()
        proto_step.name = step.name
        proto_step.guid = str(step.guid)
        proto_step.elementGuids.extend(step.element_ids)
        proto_step.elementHeld.extend(step.elements_held)
        proto_step.isBuilt = step.is_built
        proto_step.isPlanned = step.is_planned
        proto_step.actor = step.actor
        proto_step.priority = step.priority
        proto_data.steps.append(proto_step)

    return proto_data


@pb_deserializer(building_plan_pb2.BuildingPlanData)
def buildingplan_from_pb(proto_data: building_plan_pb2.BuildingPlanData) -> BuildingPlan:
    """Convert a protobuf message to a BuildingPlan object.

    Parameters
    ----------
    proto_data : building_plan_pb2.BuildingPlanData
        The protobuf message to convert.

    Returns
    -------
    BuildingPlan
        The reconstructed BuildingPlan object.
    """
    plan = BuildingPlan()
    plan.name = proto_data.name
    plan._guid = proto_data.guid

    for step in proto_data.steps:
        step_obj = Step(
            element_ids=list(step.elementGuids), elements_held=list(step.elementHeld), is_built=step.isBuilt, is_planned=step.isPlanned, actor=step.actor, priority=step.priority
        )
        step_obj.name = step.name
        step_obj._guid = step.guid
        plan.add_step(step_obj)

    return plan



import pytest

from compas.geometry import Frame
from compas.geometry import Box
from compas.geometry import Cylinder
from compas.geometry import Sphere
from compas_model.models import Model
from compas_model.elements import Element
from compas_timber.model import TimberModel
from compas_timber.elements import Beam
from compas_timber.planning import SimpleSequenceGenerator
from compas_timber.planning import BuildingPlan

from compas_timber_pb.planning import BuildingPlanModelContainer

from compas_pb import pb_dump_json
from compas_pb import pb_load_json
from compas_pb import pb_dump_bts
from compas_pb import pb_load_bts


@pytest.fixture
def timber_model():
    model = TimberModel()
    model.add_element(Beam(name="Beam01", length=4000, width=200, height=300, frame=Frame.worldXY()))
    model.add_element(Beam(name="Beam02", length=5000, width=250, height=350, frame=Frame.worldYZ()))
    model.add_element(Beam(name="Beam03", length=5000, width=250, height=350, frame=Frame.worldZX()))
    return model


@pytest.fixture
def just_model():
    model = Model()
    model.add_element(Element(name="Element01", frame=Frame.worldXY(), geometry=Box(100.0).to_mesh()))
    model.add_element(Element(name="Element02", frame=Frame.worldYZ(), geometry=Box(200.0).to_mesh()))
    model.add_element(Element(name="Element03", frame=Frame.worldYZ(), geometry=Box(300.0).to_mesh()))
    model.add_element(Element(name="Element04", frame=Frame.worldYZ(), geometry=Box(400.0).to_mesh()))
    return model


def test_building_plan_timber_model(timber_model):
    element_map = {str(element.guid): element for element in timber_model.elements()}
    plan = SimpleSequenceGenerator(timber_model).result

    geometries_map = {guid: f"{element.name}.obj" for guid, element in element_map.items()}

    json_plan = pb_dump_json({"elements": element_map, "plan": plan, "geometries": geometries_map})

    loaded_plan = pb_load_json(json_plan)

    assert "elements" in loaded_plan
    assert "plan" in loaded_plan
    assert "geometries" in loaded_plan

    for guid, element in loaded_plan["elements"].items():
        assert isinstance(element, Beam)
        assert guid in element_map
        assert guid == str(element.guid)

    assert isinstance(loaded_plan["plan"], BuildingPlan)

    for loaded_step, step in zip(loaded_plan["plan"].steps, plan.steps):
        assert step.element_ids == loaded_step.element_ids


def test_building_plan_justmodel(just_model):
    element_map = {str(element.guid): element for element in just_model.elements()}
    plan = SimpleSequenceGenerator(just_model).result

    geometries_map = {guid: f"{element.name}.obj" for guid, element in element_map.items()}

    json_plan = pb_dump_json({"elements": element_map, "plan": plan, "geometries": geometries_map})

    loaded_plan = pb_load_json(json_plan)

    assert isinstance(loaded_plan, dict)
    assert "elements" in loaded_plan
    assert "plan" in loaded_plan
    assert "geometries" in loaded_plan

    for guid, element in loaded_plan["elements"].items():
        assert isinstance(element, Element)
        assert guid in element_map
        assert guid == str(element.guid)

    assert isinstance(loaded_plan["plan"], BuildingPlan)

    for loaded_step, step in zip(loaded_plan["plan"].steps, plan.steps):
        assert step.element_ids == loaded_step.element_ids


def test_buildingplanmodelcontainer_serialization(just_model):
    """Test serialization and deserialization of BuildingPlanModelContainer."""
    element_map = {str(element.guid): element for element in just_model.elements()}
    plan = SimpleSequenceGenerator(just_model).result
    geometries_map = {guid: f"{element.name}.obj" for guid, element in element_map.items()}
    container = BuildingPlanModelContainer(plan, element_map, geometries_map)

    serialized = pb_dump_bts(container)
    loaded_container = pb_load_bts(serialized)

    assert isinstance(loaded_container, BuildingPlanModelContainer)

    assert isinstance(loaded_container.plan, BuildingPlan)
    assert loaded_container.plan.name == container.plan.name
    assert len(loaded_container.plan.steps) == len(container.plan.steps)

    assert set(loaded_container.elements.keys()) == set(container.elements.keys())
    for guid in container.elements:
        assert str(loaded_container.elements[guid].guid) == str(container.elements[guid].guid)
        assert loaded_container.elements[guid].name == container.elements[guid].name

    assert set(loaded_container.geometries.keys()) == set(container.geometries.keys())
    for guid in container.geometries:
        assert loaded_container.geometries[guid] == container.geometries[guid]


def test_buildingplanmodelcontainer_shapes(just_model):
    """Test serialization and deserialization of BuildingPlanModelContainer."""
    element_map = {str(element.guid): element for element in just_model.elements()}
    elements = list(just_model.elements())
    plan = SimpleSequenceGenerator(just_model).result
    geometries_map = {
        str(elements[0].guid): f"{elements[0].name}.obj",
        str(elements[1].guid): Box(3.0),
        str(elements[2].guid): Cylinder(2.0, 30.0),
        str(elements[3].guid): Sphere(100.0),
    }
    container = BuildingPlanModelContainer(plan, element_map, geometries_map)

    serialized = pb_dump_bts(container)
    loaded_container = pb_load_bts(serialized)

    assert isinstance(loaded_container, BuildingPlanModelContainer)

    assert isinstance(loaded_container.plan, BuildingPlan)
    assert loaded_container.plan.name == container.plan.name
    assert len(loaded_container.plan.steps) == len(container.plan.steps)

    assert set(loaded_container.elements.keys()) == set(container.elements.keys())
    for guid in container.elements:
        assert str(loaded_container.elements[guid].guid) == str(container.elements[guid].guid)
        assert loaded_container.elements[guid].name == container.elements[guid].name

    assert set(loaded_container.geometries.keys()) == set(container.geometries.keys())

    assert isinstance(loaded_container.geometries[str(elements[0].guid)], str)
    assert isinstance(loaded_container.geometries[str(elements[1].guid)], Box)
    assert isinstance(loaded_container.geometries[str(elements[2].guid)], Cylinder)
    assert isinstance(loaded_container.geometries[str(elements[3].guid)], Sphere)
